-- screen.e: access to the screen

	      --------------------------------
	      -- graphics portion of screen --
	      --------------------------------
-- in calls to read_screen and write_screen
-- the screen looks like:

-- (1,1)..................(1,HSIZE)
-- ................................
-- ................................
-- (VSIZE,1)..........(VSIZE,HSIZE)

-- Note however that pixel() uses origin (0,0) coordinates

include dos_rescue.ew as dos_rescue

global constant HSIZE = 80 * 8,  -- horizontal size (char positions)
		VSIZE = 26 * 16, -- vertical size (lines) graphics portion
	    FULLVSIZE = 30 * 16  -- vertical size (lines) full screen  
	   
	   ----------------------------
	   -- text portion of screen --
	   ----------------------------

global constant QUAD_LINE = VSIZE/16 + 1,
		WARP_LINE = VSIZE/16 + 2,
		CMD_LINE  = VSIZE/16 + 3,
		MSG_LINE  = VSIZE/16 + 4

global constant CMD_POS = 18,     -- place for first char of user command
	       WARP_POS = 9,      -- place for "WARP:" to appear
	       DREP_POS = 51,     -- place for damage report
	       WEAPONS_POS = 27,  -- place for torpedos/pos/deflectors display
	       ENERGY_POS = 67,   -- place for ENERGY display
	       DOCKING_POS = 29,  -- place for "DOCKING"
	       MSG_POS = 16,      -- place for messages to start
	       DIRECTIONS_POS = 1 -- place to put directions


global type h_coord(integer x)
-- true if x is a horizontal screen coordinate
    return x >= 1 and x <= HSIZE
end type

global type v_coord(integer y)
-- true if y is a vertical screen coordinate
    return y >= 1 and y <= VSIZE
end type

global type extended_h_coord(atom x)
    -- horizontal coordinate, can be reasonably far off the screen
    return x >= -1000 and x <= HSIZE + 1000
end type

global type extended_v_coord(atom y)
    -- vertical coordinate, can be reasonably far off the screen
    return y >= -1000 and y <= VSIZE + 1000
end type

global boolean scanon -- galaxy scan on/off

-- override position()

without warning
global procedure position(integer line, integer column)
    cursor_line = line
    cursor_column = column
end procedure

global type image(sequence s)
-- a 2-d rectangular image
    return sequence(s[1])
end type

global function is_pod(image s)
-- check if an image contains any POD pixels, range 16..31 
    sequence row
    
    for i = 1 to length(s) do
	row = s[i]
	for j = 1 to length(row) do
	    if row[j] >= 16 and row[j] <= 31 then
		return TRUE  
	    end if
	end for
    end for
    return FALSE
end function

global function all_clear(image s)
-- check if an image contains only 0's, stars and torpedo/phasor/explosion
-- "fluff".
-- > SOLID_BLACK is just "fluff" (not really there - phasors etc.) or stars 
-- SOLID_BLACK is transparent (but solid) area of an object
    sequence row
    
    for i = 1 to length(s) do
	row = s[i]
	for j = 1 to length(row) do
	    if row[j] >= 1 and row[j] <= SOLID_BLACK then
		return FALSE  
	    end if
	end for
    end for
    return TRUE
end function

global sequence screen   -- A copy of the pixels on the action screen 
screen = {} -- undefined initially

global sequence gscreen  -- A copy of the pixels on the galaxy scan
gscreen = {}
		       
global function read_screen(h_coord x, extended_v_coord y, image shape)
-- return the rectangular 2-d image at (x, y) with the same dimensions as shape
    sequence r
    integer width1
    
    width1 = length(shape[1])-1
    r = repeat(0, length(shape))
    y -= 1
    for i = 1 to length(shape) do
	r[i] = screen[y+i][x..x+width1]
    end for
    return r
end function

global function read_torp(h_coord x, v_coord y)
-- read the torpedo shape - assumed to be 2x2 for higher performance
    return {screen[y][x..x+1], 
	    screen[y+1][x..x+1]}
end function

global procedure display_image(sequence xy, sequence pixels)
   -- should really call allegro and blit a bitmap!!!
--  integer startx
    
    for i = 1 to length(pixels) do
	pixel(pixels[i], xy)
	xy[2] += 1
    end for
end procedure

global function update_image(image background, positive_int x, natural y,
			     image shape)
-- write shape into background image    
    integer width1
    
    width1 = length(shape[1])-1
    y -= 1
    for i = 1 to length(shape) do
	background[y+i][x..x+width1] = shape[i]
    end for
    return background
end function

global procedure write_screen(h_coord x, extended_v_coord y, image shape)
-- write a rectangular 2-d image into the screen variable and onto the 
-- physical screen

    integer width1
    
    width1 = length(shape[1])-1
    y -= 1
    for i = 1 to length(shape) do
	screen[y+i][x..x+width1] = shape[i]
    end for
    if not scanon then
	display_image({x-1, y}, shape)
    end if
end procedure

global procedure gwrite_screen(h_coord x, extended_v_coord y, image shape)
-- write a rectangular 2-d image into the gscreen variable and onto the 
-- physical screen

    integer width1
    
    width1 = length(shape[1])-1
    y -= 1
    for i = 1 to length(shape) do
	gscreen[y+i][x..x+width1] = shape[i]
    end for
    display_image({x-1, y}, shape)
end procedure

-- delayed-write rectangular region {{xmin,ymin}, {xmax, ymax}}
object delayed 
delayed = 0

global procedure delayed_write_screen(h_coord x, extended_v_coord y, image shape)
-- write a rectangular 2-d image into the screen variable but *not* to the 
-- physical screen (yet). Avoids flicker.

    integer w, x1, y1
    
    w = length(shape[1])-1
    y -= 1
    for i = 1 to length(shape) do
	screen[y+i][x..x+w] = shape[i]
    end for
    y += 1
    x1 = x + w
    y1 = y + length(shape)-1
    if atom(delayed) then
	delayed = {{x,y},{x1,y1}}
    else
	-- enlarge the delayed-write region?
	if x < delayed[1][1] then
	    delayed[1][1] = x
	end if
	if y < delayed[1][2] then
	    delayed[1][2] = y
	end if
	if x1 > delayed[2][1] then
	    delayed[2][1] = x1
	end if
	if y1 > delayed[2][2] then
	    delayed[2][2] = y1
	end if
    end if
end procedure

global procedure flush_screen()
-- write the delayed-write region to the physical screen   
    sequence delayed_image
    natural xmin, xmax, ybase
    
    if not scanon and sequence(delayed) then
	delayed_image = repeat(0, delayed[2][2] - delayed[1][2] + 1)
	xmin = delayed[1][1]
	xmax = delayed[2][1]
	ybase = delayed[1][2]-1
	for i = 1 to length(delayed_image) do
	    delayed_image[i] = screen[ybase+i][xmin..xmax]
	end for
	display_image(delayed[1]-1, delayed_image)
    end if
    delayed = 0
end procedure

global procedure write_torp(extended_h_coord x, v_coord y, image shape)
-- write a torpedo-shaped object - assumed to be 2x2 for higher performance
    screen[y  ][x..x+1] = shape[1]
    screen[y+1][x..x+1] = shape[2]
    if not scanon then
	x -= 1
	pixel(shape[1], {x, y-1})
	pixel(shape[2], {x, y  })
    end if
end procedure

global constant BLANK_LINE = repeat(' ', 80)

sequence black_screen
-- (doesn't use much space, because of shared pointers)
black_screen = repeat(repeat(BLACK, HSIZE), VSIZE)

-- quadrants on galaxy scan
global constant QXSIZE = 64, 
		QYSIZE = 40,
		QXBASE = 32,
		QYBASE = -8      

sequence blue_screen -- background for scan
sequence blue_black_line, blue_line

blue_screen = repeat(0, VSIZE)

blue_black_line = repeat(BLUE, QXBASE+QXSIZE) & 
		  repeat(BLACK, QXSIZE*G_SIZE) &
		  repeat(BLUE, QXBASE+QXSIZE)

blue_line = repeat(BLUE, HSIZE)

for i = 1 to VSIZE do
    if i > QYBASE+QYSIZE and i <= QYBASE+QYSIZE+G_SIZE*QYSIZE then
	blue_screen[i] = blue_black_line
    else
	blue_screen[i] = blue_line
    end if
end for

global procedure BlueScreen()
-- set physical upper screen to BLUE/BLACK for scan
    gscreen = blue_screen
    display_image({0,0}, gscreen)
end procedure

global procedure BlackScreen()
-- Set physical upper screen to BLACK
-- and set the screen variable too.
-- Initially the screen variable is undefined

    screen = black_screen
    if not scanon then
	display_image({0, 0}, screen)
    end if
end procedure

global procedure ShowScreen()
-- rewrite screen after galaxy scan
    display_image({0, 0}, screen)
end procedure

global atom frame_rate
frame_rate = 0.5   -- time between snapshots

global sequence frame_list
frame_list = {}

global boolean recording
recording = FALSE

sequence text_ops
text_ops = {}  -- list of text output commands since last snapshot

global procedure console_puts(object string)
-- write text to the screen
    sequence gpos
    
    if atom(string) then
	string = {string}
    elsif length(string) = 0 then
	return
    end if
    
    gpos = {(cursor_column-1) * 8, (cursor_line-1) * 16}
    putsxy(gpos, string, last_text_color, last_bk_color)
    if recording then
	text_ops = append(text_ops, 
		      {cursor_line, cursor_column, last_text_color,
		       last_bk_color, string})
    end if
    cursor_column += length(string)
end procedure

global procedure console_printf(sequence format, object values)
-- write printf output to the screen
    console_puts(sprintf(format, values))   
end procedure

global procedure snap()
-- take a snapshot of the graphics and text 
    if recording then
	if scanon then
	    frame_list = append(frame_list, {gscreen, text_ops}) 
	else    
	    frame_list = append(frame_list, {screen, text_ops})
	end if
	text_ops = {}
    end if
end procedure

global procedure task_video_snapshot()
-- task: take a snapshot of the screen
    
    while recording do
	-- copy the current screen variable and recent text operations
      
	snap()
	
	if length(frame_list) > 1 then
	    -- see ex.pro - do we ever fall behind?
	    if length(frame_list) > 20 then
		-- machine is too slow, can't keep up
		frame_rate *= 1.1
	    end if
	end if
      
	task_schedule(t_video_snapshot, {.9*frame_rate, frame_rate})
	task_yield()
    end while
end procedure 

constant SYNC = 88,
	 TEXT = 99,
	 RLE = 10

global integer video_file

function all_same(sequence s)
    integer x
    
    x = s[1]
    for i = 2 to length(s) do
	if x != s[i] then
	    return FALSE
	end if
    end for
    return TRUE
end function

procedure save_segment(sequence row, integer y, integer d1, integer d2)
-- save a horizontal segment of pixels to disk  
    sequence seg
    boolean same
    
    seg = row[d1..d2]
    same = length(seg) > 3 and all_same(seg)
    
    if same then
	puts(video_file, floor(y/256) + RLE)
    else
	puts(video_file, floor(y/256)) -- 0, 1, 2
    end if
    puts(video_file, y)
		    
    puts(video_file, floor(d1/256))
    puts(video_file, d1)
		    
    puts(video_file, floor(d2/256))
    puts(video_file, d2)
			
    if same then
	puts(video_file, floor(length(seg))/256)
	puts(video_file, length(seg))
	puts(video_file, seg[1])
    else
	puts(video_file, seg)
    end if
end procedure

procedure save_text(sequence ops)
-- save recent text ops to disk
    sequence op
    
    for i = 1 to length(ops) do
	op = ops[i]
	puts(video_file, TEXT) -- indicates text op
	puts(video_file, op[1..4]+2) -- avoid any 0's
	puts(video_file, op[5])
	puts(video_file, 0) -- end of string marker
    end for
end procedure

global procedure task_video_save()
-- task: compress one or more snapshots, and save to disk
-- This task is a "time share" task that gets control when
-- no other real-time tasks are ready to run
    sequence frame, previous_frame, recent_text
    integer d1, d2, same, width
    sequence previous_row, row, text

    previous_frame = repeat(repeat(0, HSIZE), VSIZE)
    
    while TRUE do
	if length(frame_list) > 0 then
	    -- one or more frames in the queue
	    frame = frame_list[1][1]
	    text = frame_list[1][2]
	    puts(video_file, SYNC)
	    frame_list = frame_list[2..$]
	    
	    width = HSIZE
	    
	    for y = 1 to VSIZE do
		-- compare this frame with the previous one and record the diffs
		row = frame[y]
		task_yield()
		
		previous_row = previous_frame[y]
		task_yield()
		
		if not equal(row, previous_row) then
		    -- this row has changed vs. previous frame
		    d1 = 1
		    while d1 <= width do
			-- skip same pixels
			while d1 <= width and 
			      row[d1] = previous_row[d1] do
			    d1 += 1
			end while
			if d1 > width then
			    exit
			end if
			
			-- we're at a different pixel
			d2 = d1
			same = 0
			while TRUE do
			    if row[d2] = previous_row[d2] then
				same += 1
				if same > 8 then -- small segments are not efficient
				    exit
				end if
			    else
				same = 0
			    end if
			    if d2 = width then
				exit
			    end if
			    d2 += 1
			end while
			d2 -= same -- back up to last different pixel
			
			-- this segment differs - save it
			save_segment(row, y, d1, d2)
			task_yield()
			
			d1 = d2+same+1
		    end while
		end if
	    end for

	    save_text(text)
	    
	    previous_frame = frame
	else
	    -- nothing to do
	    c_proc(Sleep, {1})  -- .001 sec. 
				-- (be nice to any other programs running)
	    task_yield()
	end if
    end while
end procedure

